import databases
import asyncio
import sqlalchemy
from sqlalchemy.dialects.postgresql.json import JSONB

metadata = sqlalchemy.MetaData()

geo_jsonb = sqlalchemy.Table(
    "geo_jsonb",
    metadata,
    sqlalchemy.Column("id", sqlalchemy.Integer, primary_key=True),
    sqlalchemy.Column("accession", sqlalchemy.String(length=100)),
    sqlalchemy.Column("doc", JSONB),
)

database = databases.Database(
    'postgresql://postgres:this2222@localhost:5432/omicidx',
    min_size=5,
    max_size=20)


async def main():

    await database.connect()
    stmt = geo_jsonb.select()
    res = await database.fetch_all(stmt)
    for r in res:
        print(r['accession'])
    await insert_record('abc', {'abc': 123})


async def get_saved_gses():
    async with databases.Database(
            'postgresql://postgres:this2222@localhost:5432/omicidx') as db:
        stmt = geo_jsonb.select(geo_jsonb.c.accession)
        res = [r.get(0) for r in await db.fetch_all(stmt)]
        return res


async def insert_records(vals):
    async with databases.Database(
            'postgresql://postgres:this2222@localhost:5432/omicidx') as db:
        stmt = 'insert into geo_jsonb (accession, doc) values (:accession, :doc)'

        stmt = geo_jsonb.insert()
        res = await db.execute_many(stmt, vals)


if __name__ == '__main__':
    asyncio.get_event_loop().run_until_complete(main())



"""Utilities for working with google cloud storage"""
from google.cloud import storage
import logging
from typing import Tuple


def upload_blob_to_gcs(bucket_name, source_file_name, destination_blob_name):
    """Uploads a file to the bucket.

    Parameters
    ----------
    bucket name: str
    source_file_name: str
        A local filename
    destination_blob_name: str
        The ``path`` of the object in storage
    """
    storage_client = storage.Client()
    bucket = storage_client.get_bucket(bucket_name)
    blob = bucket.blob(destination_blob_name)

    logging.info(
        f'Uploading file {source_file_name} to gs://{bucket_name}/{destination_blob_name}.'
    )

    blob.upload_from_filename(source_file_name)

    logging.info(
        f'File {source_file_name} uploaded to gs://{bucket_name}/{destination_blob_name}.'
    )


def list_blobs(bucket_name, prefix):
    """list blobs in a bucket given a prefix

    Parameters
    ----------
    bucket name: str
    prefix: str
        a `matching` string
    """
    storage_client = storage.Client()
    return storage_client.list_blobs(bucket_name, prefix=prefix)


def parse_gcs_url(url: str) -> tuple():
    import re
    splitter = re.compile('gs://([^/.]+)/(.*)')
    m = splitter.match(url)
    if m is None:
        logging.error(f'{url} does not look like a gs url')
    return ((m[1], m[2]))



"""Config comes from environment variables"""
import os


class Config(object):
    def __init__(self):

        self.ES_HOST = os.getenv('ES_HOST')
        self.GCS_STAGING_URL = os.getenv('GCS_STAGING_URL')
        self.GCS_EXPORT_URL = os.getenv('GCS_EXPORT_URL')


config = Config()



import google
import os
import logging
import json
from google.cloud import bigquery
from google.cloud import storage
from google.cloud.bigquery import SchemaField

logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')


def _get_field_schema(field):
    """convert bigquery field (dict) to SchemaField

    Works recursively if necessary to deal with 
    nested fields"""

    name = field['name']
    field_type = field.get('type', 'STRING')
    mode = field.get('mode', 'NULLABLE')
    fields = field.get('fields', [])

    if fields:
        subschema = []
        for f in fields:
            fields_res = _get_field_schema(f)
            subschema.append(fields_res)
    else:
        subschema = []

    field_schema = SchemaField(name=name,
                               field_type=field_type,
                               mode=mode,
                               fields=subschema)
    return field_schema


def parse_bq_json_schema(schema_filename):
    """Convert bigquery JSON file to python Bigquery Schema
    
    Parameters
    ----------
    schema_filename: str
        A json file with bigquery schema dump
    """
    schema = []
    with open(schema_filename, 'r') as infile:
        jsonschema = json.load(infile)

    for field in jsonschema:
        schema.append(_get_field_schema(field))

    return schema


def _load_file_to_bigquery(dataset,
                           table,
                           uri,
                           job_config: bigquery.LoadJobConfig,
                           schema=None,
                           drop: bool = True):

    client = bigquery.Client()
    dataset_id = dataset

    dataset_ref = client.dataset(dataset_id)

    if (drop):
        try:
            client.delete_table(dataset_ref.table(table))
            logging.info(f'Table {table} dropped from dataset {dataset}')
        except:
            pass

    uri = uri

    load_job = client.load_table_from_uri(
        uri,
        dataset_ref.table(table),
        location="US",  # Location must match that of the destination dataset.
        job_config=job_config,
    )  # API request
    logging.info("Starting job {}".format(load_job.job_id))
    logging.info(f"Loading {uri} to table {table}")
    try:
        load_job.result()  # Waits for table load to complete.
        logging.info("Job finished.")
    except google.api_core.exceptions.BadRequest:
        logging.error(f"Job loading {uri} into {dataset}.{table} failed.")
        logging.error(load_job.errors)

    destination_table = client.get_table(dataset_ref.table(table))
    logging.info("Loaded {} rows.".format(destination_table.num_rows))


def load_json_to_bigquery(dataset, table, uri, schema=None, drop=True):
    """Load a file from google cloud storage into BigQuery

    Parameters
    ----------
    dataset: str
        The Bigquery dataset
    table: str
        The Bigquery table
    uri: str
        The google cloud storage uri (``gs://....``)
    schema: List[SchemaField] objects or ``None``
        The schema as a list of `bigquery.SchemaField` objects
    drop: boolean
        Drop the table or not.
    """
    job_config = bigquery.LoadJobConfig()
    if (schema is not None):
        job_config.schema = schema
    else:
        job_config.autodetect = True
    job_config.source_format = bigquery.SourceFormat.NEWLINE_DELIMITED_JSON

    _load_file_to_bigquery(dataset, table, uri, job_config, schema, drop)


def load_csv_to_bigquery(dataset, table, uri, schema=None, drop=True,
                         **kwargs):
    """Load a file from google cloud storage into BigQuery

    Parameters
    ----------
    dataset: str
        The Bigquery dataset
    table: str
        The Bigquery table
    uri: str
        The google cloud storage uri (``gs://....``)
    schema: List[SchemaField] objects or ``None``
        The schema as a list of `bigquery.SchemaField` objects
    drop: boolean
        Drop the table or not.
    """
    job_config = bigquery.LoadJobConfig(**kwargs)
    if (schema is not None):
        job_config.schema = schema
    else:
        job_config.autodetect = True
    job_config.source_format = bigquery.SourceFormat.CSV
    job_config.skip_leading_rows = 1  #avoid header

    _load_file_to_bigquery(dataset, table, uri, job_config, schema, drop)


def copy_table(src_dataset: str,
               dest_dataset: str,
               src_table: str,
               dest_table: str,
               drop=True):
    client = bigquery.Client()
    source_dataset = client.dataset(src_dataset)
    source_table_ref = source_dataset.table(src_table)

    dest_dataset = client.dataset(dest_dataset)
    dest_table_ref = dest_dataset.table(dest_table)

    if (drop):
        try:
            client.delete_table(dest_table_ref)
            logging.info(
                f'Table {dest_table} dropped from dataset {dest_dataset}')
        except:
            pass
    job = client.copy_table(
        source_table_ref,
        dest_table_ref,
        # Location must match that of the source and destination tables.
        location="US",
    )  # API request

    logging.info("Starting job {}".format(job.job_id))
    logging.info(
        f"copying {src_dataset}.{src_table} to {dest_dataset}.{dest_table}")

    try:
        job.result()  # Waits for table load to complete.
        logging.info("Job finished.")
    except google.api_core.exceptions.BadRequest:
        logging.error(
            f"Job copying {src_dataset}.{src_table} to {dest_dataset}.{dest_table} failed"
        )
        logging.error(load_job.errors)


def query_to_destination(dest_dataset: str,
                         dest_table: str,
                         sql: str,
                         drop=True):
    client = bigquery.Client()
    dest_dataset = client.dataset(dest_dataset)
    dest_table_ref = dest_dataset.table(dest_table)

    if (drop):
        try:
            client.delete_table(dest_table_ref)
            logging.info(
                f'Table {dest_table} dropped from dataset {dest_dataset}')
        except:
            pass

    job_config = bigquery.QueryJobConfig()
    job_config.destination = dest_table_ref

    # Start the query, passing in the extra configuration.
    query_job = client.query(
        sql,
        # Location must match that of the dataset(s) referenced in the query
        # and of the destination table.
        location="US",
        job_config=job_config,
    )

    logging.info("Starting job {}".format(query_job.job_id))
    logging.info(f"----------------------{sql}---------------------")
    logging.info(f"querying to {dest_dataset}.{dest_table}")

    try:
        query_job.result()  # Waits for table load to complete.
        logging.info("Query results loaded to table {}".format(
            dest_table_ref.path))
    except google.api_core.exceptions.BadRequest:
        logging.error(f"querying to {dest_dataset}.{dest_table} failed")
        logging.error(query_job.errors)


def query(sql: str):
    client = bigquery.Client()

    job_config = bigquery.QueryJobConfig()

    # Start the query, passing in the extra configuration.
    query_job = client.query(
        sql,
        # Location must match that of the dataset(s) referenced in the query
        # and of the destination table.
        location="US",
        job_config=job_config,
    )

    logging.info("Starting job {}".format(query_job.job_id))
    logging.info(f"----------------------{sql}---------------------\n")

    try:
        query_job.result()  # Waits for table load to complete.
        logging.info("Query completed")
    except google.api_core.exceptions.BadRequest:
        logging.error(f"query failed")
        logging.error(query_job.errors)


def table_to_gcs(dataset: str,
                 table: str,
                 uri: str,
                 gzip: bool = True,
                 delete_first: bool = True):
    """Load a file from google cloud storage into BigQuery

    Parameters
    ----------
    dataset: str
        The Bigquery dataset
    table: str
        The Bigquery table
    uri: str
        The google cloud storage uri (``gs://....``)
    gzip: bool
        Compress output with gzip or not
    """
    client = bigquery.Client()

    destination_uri = uri
    dataset_ref = client.dataset(dataset)
    table_ref = dataset_ref.table(table)

    logging.info("Exporting {}.{} to {}".format(dataset, table,
                                                destination_uri))

    job_config = bigquery.ExtractJobConfig()

    if (gzip):
        job_config.compression = bigquery.Compression.GZIP
        job_config.destination_format = bigquery.DestinationFormat.NEWLINE_DELIMITED_JSON

    extract_job = client.extract_table(
        table_ref,
        destination_uri,
        # Location must match that of the source table.
        location="US",
        job_config=job_config)  # API request
    try:
        extract_job.result()  # Waits for table load to complete.
        logging.info("Extract completed")
    except google.api_core.exceptions.BadRequest:
        logging.error(f"extract failed")
        logging.error(extract_job.errors)

    logging.info("Exported {}.{} to {}".format(dataset, table,
                                               destination_uri))



#!/usr/bin/env python
import click
import subprocess
import logging
import os
import omicidx
from .config import config
from .bigquery_utils import *
import omicidx.sra.parser
from .click_root import cli
# from .geo_cli import geo
import json
from xml.etree import ElementTree as et
from .utils import dateconverter
from .gcs_utils import upload_blob_to_gcs, parse_gcs_url

logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s')
logger = logging.getLogger(__name__)


@cli.group(help="Use these commands to process SRA metadata")
def sra():
    pass


@sra.command("download",
             help="""Downloads the files necessary to build
             the SRA json conversions of the XML files.

             Files will be placed in the <mirrordir> directory. Mirrordirs
             have the format `NCBI_SRA_Mirroring_20190801_Full`.
             """)
@click.argument('mirrordir')
def download_mirror_files(mirrordir):
    logger.info('getting xml files')
    subprocess.run(
        "wget -nH -nv -np --cut-dirs=3 -r -e robots=off {}/{}/".format(
            "http://ftp.ncbi.nlm.nih.gov/sra/reports/Mirroring", mirrordir),
        shell=True)

    logger.info('getting SRA Accessions file')
    subprocess.run(
        "wget -nv ftp://ftp.ncbi.nlm.nih.gov/sra/reports/Metadata/SRA_Accessions.tab -P {}"
        .format(mirrordir),
        shell=True)


@sra.command("parse-entity",
             help="""SRA XML to JSON

             Transforms an SRA XML mirroring metadata file into
             corresponding JSON format files. JSON is line-delimited
             JSON (not an array).""")
@click.option('--entity', '-e', help="entity to process")
def process_xml_entity(entity):

    fname = "meta_{}_set.xml.gz".format(entity)

    entity = entity

    parsers = {
        'study': omicidx.sra.parser.SRAStudyRecord,
        'sample': omicidx.sra.parser.SRASampleRecord,
        'run': omicidx.sra.parser.SRARunRecord,
        'experiment': omicidx.sra.parser.SRAExperimentRecord
    }
    sra_parser = parsers[entity]

    import omicidx.sra.pydantic_models as p
    parsers = {
        'study': p.SraStudy,
        'sample': p.SraSample,
        'run': p.SraRun,
        'experiment': p.SraExperiment
    }
    pydantic_model = parsers[entity]

    logger.info('using {} entity type'.format(entity))
    logger.info('parsing {} records'.format(entity))
    n = 0
    outfname = "{}.json".format(entity)
    ENTITY = entity.upper()
    with open(outfname, 'w') as outfile:
        with omicidx.sra.parser.open_file(fname) as f:
            for event, element in et.iterparse(f):
                if (event == 'end' and element.tag == ENTITY):
                    rec = sra_parser(element).data
                    n += 1
                    if ((n % 100000) == 0):
                        logger.info('parsed {} {} entries'.format(entity, n))
                    outfile.write(
                        json.dumps(pydantic_model(**rec).dict(),
                                   default=dateconverter) + "\n")
                    element.clear()
            logger.info('parsed {} entity entries'.format(n))


@sra.command('upload', help="""Upload SRA json to GCS""")
@click.argument('mirrordir')
def upload_processed_sra_data(mirrordir):

    (bucket, path) = parse_gcs_url(config.GCS_STAGING_URL)
    for entity in 'study sample experiment run'.split():
        fname = entity + '.json'
        loc_fname = os.path.join(mirrordir, fname)
        upload_blob_to_gcs(bucket, loc_fname, os.path.join(path, fname))

    fname = 'SRA_Accessions.tab'
    loc_fname = os.path.join(mirrordir, fname)
    upload_blob_to_gcs(bucket, loc_fname, os.path.join(path, fname))


@sra.command(help="""Load gcs files to Bigquery""")
def load_sra_data_to_bigquery():
    from importlib import resources

    (bucket, path) = parse_gcs_url(config.GCS_STAGING_URL)
    for i in 'study sample experiment run'.split():
        with resources.path('omicidx_builder.data.bigquery_schemas',
                            f"{i}.schema.json") as schemafile:
            load_json_to_bigquery('omicidx_etl',
                                  f'sra_{i}',
                                  os.path.join(config.GCS_STAGING_URL,
                                               f'{i}.json'),
                                  schema=parse_bq_json_schema(schemafile))

    with resources.path('omicidx_builder.data.bigquery_schemas',
                        f"sra_accession.schema.json") as schemafile:
        load_csv_to_bigquery('omicidx_etl',
                             'sra_accessions',
                             os.path.join(config.GCS_STAGING_URL,
                                          f'SRA_Accessions.tab'),
                             field_delimiter='\t',
                             null_marker='-',
                             quote_character="",
                             schema=parse_bq_json_schema(schemafile))


@sra.command(help="""ETL query to public schema for all SRA entities""")
def sra_to_bigquery():
    sql = """CREATE OR REPLACE TABLE `isb-cgc-01-0006.omicidx.sra_run` AS
SELECT 
  run.* EXCEPT (published, lastupdate, received, total_spots, total_bases, avg_length, run_date),
  CAST(acc.Updated as DATETIME) as lastupdate,
  CAST(acc.Published as DATETIME) as published,
  CAST(acc.Received as DATETIME) as received,
  CAST(run_date as DATETIME) as run_date,
  CAST(acc.Spots as INT64) as total_spots,
  CAST(acc.Bases as INT64) as total_bases,
  CAST(acc.Bases AS NUMERIC)/CAST(acc.Spots AS NUMERIC) as avg_length,
  acc.Sample as sample_accession,
  acc.Study as study_accession
FROM 
    `isb-cgc-01-0006.omicidx_etl.sra_run` run
  JOIN 
    `isb-cgc-01-0006.omicidx_etl.sra_accessions` acc
  ON acc.Accession = run.accession;
"""
    query(sql)

    sql = """CREATE OR REPLACE TABLE `isb-cgc-01-0006.omicidx.sra_experiment` AS
SELECT 
  expt.* EXCEPT (published, lastupdate, received),
  CAST(acc.Updated as DATETIME) as lastupdate,
  CAST(acc.Published as DATETIME) as published,
  CAST(acc.Received as DATETIME) as received
FROM 
    `isb-cgc-01-0006.omicidx_etl.sra_experiment` expt
  JOIN 
    `isb-cgc-01-0006.omicidx_etl.sra_accessions` acc
  ON acc.Accession = expt.accession;
    """
    query(sql)

    sql = """CREATE OR REPLACE TABLE `isb-cgc-01-0006.omicidx.sra_sample` AS
SELECT 
  sample.* EXCEPT (published, lastupdate, received),
  CAST(acc.Updated as DATETIME) as lastupdate,
  CAST(acc.Published as DATETIME) as published,
  CAST(acc.Received as DATETIME) as received,
  acc.Study as study_accession
FROM 
    `isb-cgc-01-0006.omicidx_etl.sra_sample` sample
  JOIN 
    `isb-cgc-01-0006.omicidx_etl.sra_accessions` acc
  ON acc.Accession = sample.accession;
    """
    query(sql)

    sql = """CREATE OR REPLACE TABLE `isb-cgc-01-0006.omicidx.sra_study` AS
WITH stat_agg AS (
SELECT
  COUNT(a.Accession) as sample_count,
  a.Study as study_accession
FROM 
  `isb-cgc-01-0006.omicidx_etl.sra_accessions` a
WHERE
  a.Type='SAMPLE'
GROUP BY 
  study_accession)
SELECT 
  study.* EXCEPT (published, lastupdate, received),
  CAST(acc.Updated as DATETIME) as lastupdate,
  CAST(acc.Published as DATETIME) as published,
  CAST(acc.Received as DATETIME) as received
FROM 
    `isb-cgc-01-0006.omicidx_etl.sra_study` study
  JOIN 
    `isb-cgc-01-0006.omicidx_etl.sra_accessions` acc
  ON acc.Accession = study.accession
  LEFT OUTER JOIN
    stat_agg on stat_agg.study_accession=study.accession;
    """
    query(sql)


def _sra_bigquery_for_elasticsearch():
    sql = """CREATE OR REPLACE TABLE omicidx_etl.sra_experiment_for_es AS
SELECT
  expt.*,
  STRUCT(samp).samp as sample,
  STRUCT(study).study as study
FROM 
  `isb-cgc-01-0006.omicidx.sra_experiment` expt 
  LEFT OUTER JOIN 
  `isb-cgc-01-0006.omicidx.sra_sample` samp
  ON samp.accession = expt.sample_accession
  LEFT OUTER JOIN 
  `isb-cgc-01-0006.omicidx.sra_study` study
  ON study.accession = expt.study_accession
WHERE samp.accession IS NOT NULL
AND study.accession IS NOT NULL
AND expt.study_accession is NOT NULL
AND expt.sample_accession is NOT NULL
    """
    query(sql)

    sql = """CREATE OR REPLACE TABLE omicidx_etl.sra_run_for_es AS
SELECT
  run.*,
  STRUCT(expt).expt as experiment,
  STRUCT(samp).samp as sample,
  STRUCT(study).study as study
FROM 
  `isb-cgc-01-0006.omicidx.sra_run` run
  LEFT OUTER JOIN
  `isb-cgc-01-0006.omicidx.sra_experiment` expt
  ON run.experiment_accession = expt.accession
  LEFT OUTER JOIN 
  `isb-cgc-01-0006.omicidx.sra_sample` samp
  ON samp.accession = expt.sample_accession
  LEFT OUTER JOIN 
  `isb-cgc-01-0006.omicidx.sra_study` study
  ON study.accession = expt.study_accession
WHERE samp.accession IS NOT NULL
AND study.accession IS NOT NULL
AND expt.study_accession is NOT NULL
AND expt.sample_accession is NOT NULL
AND run.experiment_accession IS NOT NULL
"""
    query(sql)

    sql = """CREATE OR REPLACE TABLE omicidx_etl.sra_sample_for_es AS
with agg_counts as 
(SELECT
  sample.accession,
  COUNT(DISTINCT(expt.accession)) as experiment_count,
  COUNT(DISTINCT(run.accession)) as run_count,
  SUM(CAST(run.total_bases as INT64)) as total_bases,
  SUM(CAST(run.total_spots as INT64)) as total_spots,
  AVG(CAST(run.total_bases as INT64)) as mean_bases_per_run
from `isb-cgc-01-0006.omicidx.sra_experiment` expt
JOIN `isb-cgc-01-0006.omicidx.sra_run` run
  ON run.experiment_accession = expt.accession and run.experiment_accession is not null and expt.accession is not null
JOIN `isb-cgc-01-0006.omicidx.sra_sample` sample
  ON expt.sample_accession = sample.accession and expt.sample_accession is not null and sample.accession is not null
JOIN `isb-cgc-01-0006.omicidx.sra_study` study
  ON expt.study_accession = study.accession and expt.study_accession is not null and study.accession is not null
group by sample.accession
)
select 
  sample.*,
  STRUCT(study).study,
  agg_counts.* except(accession)
from `isb-cgc-01-0006.omicidx.sra_experiment` expt
JOIN `isb-cgc-01-0006.omicidx.sra_sample` sample
  ON expt.sample_accession = sample.accession and expt.sample_accession is not null and sample.accession is not null
JOIN `isb-cgc-01-0006.omicidx.sra_study` study
  ON expt.study_accession = study.accession and expt.study_accession is not null and study.accession is not null
join agg_counts on agg_counts.accession=sample.accession
"""
    query(sql)

    sql = """CREATE OR REPLACE TABLE omicidx_etl.sra_study_for_es AS
with agg_counts as 
(SELECT
  study.accession,
  COUNT(DISTINCT(sample.accession)) as sample_count,
  COUNT(DISTINCT(expt.accession)) as experiment_count,
  COUNT(DISTINCT(run.accession)) as run_count,
  SUM(CAST(run.total_bases as INT64)) as total_bases,
  SUM(CAST(run.total_spots as INT64)) as total_spots,
  AVG(CAST(run.total_bases as INT64)) as mean_bases_per_run
from `isb-cgc-01-0006.omicidx.sra_experiment` expt
JOIN `isb-cgc-01-0006.omicidx.sra_run` run
  ON run.experiment_accession = expt.accession and run.experiment_accession is not null and expt.accession is not null
JOIN `isb-cgc-01-0006.omicidx.sra_sample` sample
  ON expt.sample_accession = sample.accession and expt.sample_accession is not null and sample.accession is not null
JOIN `isb-cgc-01-0006.omicidx.sra_study` study
  ON expt.study_accession = study.accession and expt.study_accession is not null and study.accession is not null
group by study.accession
)
select 
  study.*,
  agg_counts.* except(accession)
from `isb-cgc-01-0006.omicidx.sra_study` study
left join agg_counts on agg_counts.accession=study.accession
"""
    query(sql)


@sra.command(help="""ETL queries to create elasticsearch tables in bigquery""")
def sra_bigquery_for_elasticsearch():
    _sra_bigquery_for_elasticsearch()


def _sra_gcs_to_elasticsearch(entity):
    from omicidx_builder.elasticsearch_utils import (bulk_index_from_gcs,
                                                     index_for_alias,
                                                     create_alias,
                                                     delete_index)
    import uuid
    e = str(uuid.uuid4())

    (bucket, path) = parse_gcs_url(config.GCS_EXPORT_URL)
    path = path.strip('/')

    idx_name = 'sra_' + entity + '-' + e
    logger.info(f"creating index {idx_name}")

    bulk_index_from_gcs(bucket,
                        os.path.join(path, 'sra/{}-'.format(entity)),
                        idx_name,
                        id_field='accession')
    old_index = index_for_alias('sra_' + entity)
    if old_index is None:
        create_alias('sra_' + entity, idx_name)
    else:
        for index in old_index.keys():
            delete_index(index)
            logger.info(f'deleted old index {index}')
        create_alias('sra_' + entity, idx_name)
    logger.info(f'alias sra_{entity} now points to {idx_name}')


@sra.command('gcs-to-elasticsearch',
             help="""ETL query to public schema for all SRA entities""")
@click.option(
    '--entity',
    '-e',
    multiple=True,
    help="Entity (study, sample, experiment, run). Multiple values can be used",
    default=['study', 'sample', 'experiment', 'run'])
def sra_gcs_to_elasticsearch(entity):
    for e in entity:
        _sra_gcs_to_elasticsearch(e)


def _sra_to_gcs_for_elasticsearch():
    from google.cloud import storage
    (bucket, path) = parse_gcs_url(config.GCS_EXPORT_URL)
    path = path.strip('/')
    client = storage.Client()
    blobs = client.list_blobs(bucket, prefix=os.path.join(path, 'sra'))
    for blob in blobs:
        logger.info(f"deleting {blob.name}")
        blob.delete()
    for entity in 'experiment study sample run'.split():
        table_to_gcs('omicidx_etl', f'sra_{entity}_for_es',
                     f'gs://{bucket}/{path}/sra/{entity}-*.json.gz')


@sra.command("""gcs-dump""",
             help="Write json.gz format of sra entities to gcs")
def sra_to_gcs_for_elasticsearch():
    _sra_to_gcs_for_elasticsearch()


######################
# Biosample handling #
######################


@cli.group(help="Use these commands to process biosample records.")
def biosample():
    pass


from omicidx.biosample import BioSampleParser


def biosample_to_json(biosample_file: str, output: click.File):
    n = 0
    logger.info(f'starting biosample record parsing')
    for i in BioSampleParser(biosample_file):
        n += 1
        if (i is None):
            break
        if (n % 100000 == 0):
            logger.info(f'{n} biosample records parsed')
        output.write(i.as_json() + "\n")
    logger.info(f'completing biosample record parsing')


def download_biosample():
    subprocess.run(
        "wget -nv ftp://ftp.ncbi.nlm.nih.gov/biosample/biosample_set.xml.gz",
        shell=True)


def upload_biosample():
    fname = 'biosample.json'
    (bucket, path) = parse_gcs_url(config.GCS_STAGING_URL)
    upload_blob_to_gcs(bucket, fname, os.path.join(path, fname))


def load_biosample_from_gcs_to_bigquery():

    load_json_to_bigquery(
        'omicidx_etl', 'biosample',
        os.path.join(config.GCS_STAGING_URL, 'biosample.json'))


@biosample.command("""download""",
                   help="Download biosample xml file from NCBI")
def download():
    download_biosample()


@biosample.command("""upload""", help="Download biosample xml file from NCBI")
def upload():
    upload_biosample()


@biosample.command("""parse""", help="Parse xml to json, output to stdout")
@click.argument('biosample_file')
@click.argument('output', type=click.File('w'))
def to_json(biosample_file, output):
    biosample_to_json(biosample_file, output)


@biosample.command("""load""",
                   help="Load the gcs biosample.json file to bigquery")
def load_biosample_to_bigquery():
    load_biosample_from_gcs_to_bigquery()


@biosample.command("""etl-to-public""",
                   help="ETL process (copy) from etl schema to public")
def biosample_to_public():
    copy_table('omicidx_etl', 'omicidx', 'biosample', 'biosample')


@biosample.command("""gcs-dump""",
                   help="Write json.gz format of biosample to gcs")
def biosample_to_gcs():
    table_to_gcs('omicidx', 'biosample',
                 os.path.join(config.GCS_EXPORT_URL, 'biosample-*.json.gz'))


def _biosample_gcs_to_elasticsearch():
    from omicidx_builder.elasticsearch_utils import (bulk_index_from_gcs,
                                                     swap_indices_behind_alias,
                                                     index_for_alias,
                                                     create_alias,
                                                     delete_index)
    import uuid
    e = str(uuid.uuid4())

    idx_name = 'biosample-' + e
    logger.info(f"creating index {idx_name}")
    (bucket, path) = parse_gcs_url(config.GCS_STAGING_URL)
    bulk_index_from_gcs(bucket,
                        os.path.join(path, 'biosample-'),
                        idx_name,
                        max_retries=3,
                        chunk_size=2000,
                        id_field='accession')
    old_index = index_for_alias('biosample')
    if old_index is None:
        create_alias('biosample', idx_name)
        logger.info(f'alias biosample now points to {idx_name}')
    else:
        swap_indices_behind_alias('biosample', old_index, idx_name)
        logger.info(f'alias biosample now points to {idx_name}')
        delete_index(old_index)
        logger.info(f'deleted old index {old_index}')

    from omicidx_builder.elasticsearch_utils import bulk_index_from_gcs


@biosample.command("gcs-to-elasticsearch")
def biosample_gcs_to_elasticsearch():
    _biosample_gcs_to_elasticsearch()


if __name__ == '__main__':
    cli()



"""Create a basic Click instance. GEO, SRA, and Biosample 
groups then import `cli` from here to add groups"""
import click


@click.group(help="""Command-line interface for omicidx processing
""")
def cli():
    pass



import elasticsearch
from elasticsearch.helpers import bulk
from elasticsearch_dsl import connections
from omicidx_builder.config import config
import json
import logging
import gzip
logging.basicConfig(level=logging.INFO, format=logging.BASIC_FORMAT)
import os


def init_connection_object():
    connections.create_connection(alias='default',
                                  hosts=config.ES_HOST,
                                  retry_on_timeout=True,
                                  max_retries=3,
                                  timeout=30)


init_connection_object()


def get_client() -> elasticsearch.client:
    return connections.get_connection()


client = get_client()


def prep_data(fname, index, id_field):
    with gzip.open(fname) as f:
        for line in f:
            d = json.loads(line)
            d['_index'] = index
            if (id_field is not None):
                if (id_field in d):
                    d['_id'] = d[id_field]
                else:
                    continue
            yield (d)


def bulk_index(fname, index, id_field=None, **kwargs):
    bulk(get_client(), prep_data(fname, index, id_field), **kwargs)


def put_template(template_name: str, template_body: dict):
    client.indices.put_template(template_name, template_body)


def bulk_index_from_gcs(bucket, prefix, index, id_field=None, **kwargs):
    """Perform bulk indexing from a set of gcs blobs

    Parameters
    ----------
    bucket: str
        GCS bucket name
    prefix: str
        The prefix string (without wildcard) to get the right blobs
    index: str
        The elasticsearch index name
    id_field: str
        The id field name (default None) that will be used as the
        `_id` field in elasticsearch
    """
    from tempfile import NamedTemporaryFile
    from .gcs_utils import list_blobs
    logging.info(f'{bucket}/{prefix}')
    flist = list(list_blobs(bucket, prefix))
    logging.info(f'Found {len(flist)} files for indexing')
    for i in flist:
        tmpfile = NamedTemporaryFile()
        if (i.name.endswith('.gz')):
            tmpfile = NamedTemporaryFile(suffix='.gz')
        logging.info('Downloading ' + i.name)
        i.download_to_filename(tmpfile.name)
        logging.info('Indexing ' + i.name)
        bulk_index(tmpfile.name, index, id_field=id_field, **kwargs)
        tmpfile.close()


def delete_index(index):
    """Delete the named index"""
    if client.indices.exists(index):
        client.indices.delete(index)
    else:
        logging.warn(f"index {index} does not exist, so not deleted!")


def create_alias(alias: str, index: str):
    """Create an alias for an index

    Parameters
    ----------
    alias: str
        the alias name
    index: str
        the index name
    """
    if client.indices.exists(index):
        client.indices.put_alias(index, alias)
        logging.info(f"alias {alias} for index {index} created")
    else:
        logging.warn(
            f"index {index} does not exist, so alias {alias} not created")


def delete_alias(alias: str, index: str = None):
    """Create an alias for an index

    Parameters
    ----------
    alias: str
        the alias name
    index: str
        the index name
    """
    if client.indices.exists_alias(index, alias):
        client.indices.delete_alias(index, alias)
        logging.info(f"alias {alias} for index {index} deleted")
    else:
        logging.warn(f"alias {alias} does not exist")


def swap_indices_behind_alias(alias: str, old_index: str, new_index: str):
    """Swap the alias for old_index to point to new_index

    This just swaps the alias, nothing more. If the new index
    does not exist, this function is a no-op.

    Parameters
    ----------
    alias: str
        The alias to re-assign

    old_index: str
        The name of the old index. This index remains intact; only 
        the alias changes.
    
    new_index: str
        The name of the new index to which to assign the alias.
    """
    if client.indices.exists(index=new_index):
        for index in index_for_alias(alias).keys():
            delete_index(index)
        client.indices.put_alias(index=new_index, name=alias)
        logging.info(f"alias {alias} for index {new_index} created")
    else:
        logging.warn(
            f"new index {new_index} does not exist, so alias {alias} remains unchanged"
        )


def index_for_alias(alias: str):
    """get the name of the index for a given alias

    Parameters
    ----------
    alias: str
       The name of the alias

    Returns
    -------
    str: the name of the matching index (or None)
    """
    res = None
    try:
        res = client.indices.get_alias(name=alias)
    except:
        return None
    return res



"""Various utility functions for cli"""
import datetime


def dateconverter(o):
    """Convert an object from datetime to string

    Use this in the context of json to do custom
    conversions."""

    if isinstance(o, (datetime.datetime)):
        return o.__str__()



from .click_root import cli
import click
import omicidx.geo.parser as gm
import logging
logger = logging.getLogger()
from .rdbms import insert_records, get_saved_gses
import asyncio


@cli.group('geo')
def geo():
    pass


@geo.command('get-series',
             help='print out series accessions between specified dates')
@click.option('--from_date',
              '-f',
              default='2005-01-01',
              help='format as 2005-01-01')
@click.option('--to_date',
              '-t',
              default='2030-01-01',
              help='format as 2030-01-01')
def get_series(from_date, to_date):
    n = 0
    from_date = from_date.replace('-', '/')
    to_date = to_date.replace('-', '/')

    for g in gm.get_geo_accessions(add_term=f"{from_date}:{to_date}[PDAT]",
                                   etyp="GSE"):
        n += 1
        print(g)
        if (n % 1000 == 0):
            logger.info('sent {} records'.format(n))


async def _insert_gse_to_db(gse):
    vals = []
    avail_gses = get_saved_gses()
    if (gse in avail_gses):
        return True
    for z in gm.geo_soft_entity_iterator(gm.get_geo_accession_soft(gse)):
        vals.append({'accession': z.accession, 'doc': z.json()})
    await insert_records(vals)


@geo.command('gse-to-json')
@click.argument('gse')
def gse_to_json(gse):
    asyncio.run(_insert_gse_to_db(gse))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['omicidx_builder',
 'omicidx_builder.data',
 'omicidx_builder.data.bigquery_schemas']

package_data = \
{'': ['*']}

install_requires = \
['Click',
 'asyncpg>=0.20.0,<0.21.0',
 'databases>=0.2.6,<0.3.0',
 'elasticsearch>=7,<8',
 'elasticsearch_dsl>=7,<8',
 'google-cloud-bigquery',
 'google-cloud-pubsub',
 'google-cloud-storage',
 'kubernetes>=10.0.1,<11.0.0',
 'omicidx==0.3.10',
 'psycopg2>=2.8,<3.0',
 'pydantic>=1.1,<2.0',
 'sqlalchemy>=1.3,<2.0',
 'toml>=0.10,<0.11',
 'ujson>=1.35,<2.0']

entry_points = \
{'console_scripts': ['omicidx_builder = omicidx_builder.cli:cli']}

setup_kwargs = {
    'name': 'omicidx-builder',
    'version': '0.5.1',
    'description': 'Tooling to build and deploy omicidx data and resources',
    'long_description': None,
    'author': 'Sean Davis',
    'author_email': 'seandavi@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/seandavi/omicidx-builder',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""
Creates, updates, and deletes a job object.
"""

from os import path
import os

import yaml
import uuid

from kubernetes import client, config

JOB_NAME = 'omicidx-biosample-j' + str(uuid.uuid4().hex)[0:5]


def create_job_object():
    # Configureate Pod template container

    # volume inside of which you put your service account
    # This is the secret name, basically
    volume_name = "google-cloud-json"
    google_app_credentials_path = os.environ.get(
        'GOOGLE_APPLICATION_CREDENTIALS')

    # create a volume mount
    volume_mount = client.V1VolumeMount(mount_path='/etc/stuff',
                                        name=volume_name)

    # Create environment variables for container.
    # In this case, grab the values from the execution environment
    # perhaps using something like a .env file.
    env = [
        client.V1EnvVar(
            name='GOOGLE_APPLICATION_CREDENTIALS',
            # note this is a path + the filename of the app creds in the secret
            value='/etc/stuff/key.json'),  # google_app_credentials_path),
        client.V1EnvVar(name='GCS_STAGING_URL',
                        value=os.environ.get('GCS_STAGING_URL')),
        client.V1EnvVar(name='GCS_EXPORT_URL',
                        value=os.environ.get('GCS_EXPORT_URL')),
        client.V1EnvVar(name='ES_HOST', value=os.environ.get('ES_HOST'))
    ]

    # Create a volume.
    # This will go into the spec section of the template
    # Note that this specifies a secret volume
    # The secret needs to be created separately.
    volume = client.V1Volume(
        name=volume_name,
        secret=client.V1SecretVolumeSource(secret_name='google-cloud-json'))

    # Create the container section that will
    # go into the spec section
    container = client.V1Container(
        name="omicidx-builder",
        image="seandavi/omicidx-builder",
        volume_mounts=[volume_mount],
        env=env,
        command=['/bin/bash', '/code/biosample_pipeline.sh'])

    # Create and configurate a spec section
    template = client.V1PodTemplateSpec(
        metadata=client.V1ObjectMeta(labels={"app": "omicidx-builder"}),
        spec=client.V1PodSpec(restart_policy="Never",
                              volumes=[volume],
                              containers=[container]))
    # Create the specification of deployment
    spec = client.V1JobSpec(template=template, backoff_limit=4)
    # Instantiate the job object
    job = client.V1Job(api_version="batch/v1",
                       kind="Job",
                       metadata=client.V1ObjectMeta(name=JOB_NAME),
                       spec=spec)

    return job


def create_job(api_instance, job):
    api_response = api_instance.create_namespaced_job(body=job,
                                                      namespace="default")
    print("Job created. status='%s'" % str(api_response.status))


# this is just here for completeness and is not used
def update_job(api_instance, job):
    # Update container image
    job.spec.template.spec.containers[0].image = "perl"
    api_response = api_instance.patch_namespaced_job(name=JOB_NAME,
                                                     namespace="default",
                                                     body=job)
    print("Job updated. status='%s'" % str(api_response.status))


def delete_job(api_instance):
    api_response = api_instance.delete_namespaced_job(
        name=JOB_NAME,
        namespace="default",
        body=client.V1DeleteOptions(propagation_policy='Foreground',
                                    grace_period_seconds=5))
    print("Job deleted. status='%s'" % str(api_response.status))


def main():
    # Configs can be set in Configuration class directly or using helper
    # utility. If no argument provided, the config will be loaded from
    # default location.
    config.load_kube_config()
    batch_v1 = client.BatchV1Api()
    # Create a job object with client-python API. The job we
    # created is same as the `pi-job.yaml` in the /examples folder.
    job = create_job_object()

    create_job(batch_v1, job)

    print(job)

    #update_job(batch_v1, job)

    #delete_job(batch_v1)


if __name__ == '__main__':
    main()
